﻿
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using SimpleJSON;
using System.Linq;
using MVR.FileManagementSecure;
using OldMoatGames;

/* /////////////////////////////////////////////////////////////////////////////////////////////////
based on MacGruber's utils

Licensed under CC BY-SA after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea.DecalSystem
{
    public static class Utils
    {
        public const string version = "16";

        private static string[] validFileTypes = new string[] { ".asf", ".avi", ".dv", ".m4v", ".mov", ".mp4", ".mpg", ".mpeg", ".ogv", ".vp8", ".webm", ".wmv" };

        public static void ActiveSliderUI(Slider slider, bool isActive)
        {
            slider.enabled = isActive;

            if (!isActive)
                slider.transform.parent.Find("Panel").GetComponent<Image>().color = Color.gray;
            else
                slider.transform.parent.Find("Panel").GetComponent<Image>().color = Color.white;
        }

        public static IEnumerator WaitTimeCall(float time, Action callback)
        {
            yield return new WaitForSeconds(time);
            if (callback != null) callback();
        }

        public static bool IsVideoSupport(string url)
        {
            if (!url.Contains('.')) return false;
            string enfix = url.Substring(url.LastIndexOf('.'));
            //Debug.Log(enfix);
            enfix = enfix.ToLower();
            return validFileTypes.Contains(enfix);
        }

        public static void GetAllSubFiles(string dir, ref List<string> files)
        {
            if (FileManagerSecure.DirectoryExists(dir))
            {
                if (FileManagerSecure.IsDirectoryInPackage(dir))
                {
                    files = SuperController.singleton.GetFilesAtPath(dir).ToList();
                }
                else
                {
                    string[] fs = FileManagerSecure.GetFiles(dir);
                    if (fs.Length > 0)
                        foreach (var item in fs)
                        {
                            files.Add(item);
                        }

                    string[] ds = FileManagerSecure.GetDirectories(dir);
                    if (ds.Length > 0)
                    {
                        foreach (var item in ds)
                        {
                            files.Add(item);
                            GetAllSubFiles(item, ref files);
                        }
                    }
                }
            }
            else
            {
                Debug.LogError("不存在这个目录");
            }
        }

        public static Camera CenterCamera
        {
            get
            {
                if (Utils.IsInVR())
                {
                    if (SuperController.singleton.isOpenVR)
                    {
                        return SuperController.singleton.ViveCenterCamera;
                    }
                    return SuperController.singleton.OVRCenterCamera;
                }
                else
                {
                    return SuperController.singleton.MonitorCenterCamera;
                }
            }
        }

        public static string GetGameRootDir()
        {
            string dataPath = Application.dataPath;
            return dataPath.Substring(0, dataPath.LastIndexOf('/'));
        }

        public static string GetFolderFromUrl(string url)
        {
            if (url.Contains("/"))
            {
                return url.Substring(0, url.LastIndexOf("/"));
            }

            return url;
        }

        public static string GetPath(Transform current)
        {
            if (current.parent == null)
                return "/" + current.name;
            return GetPath(current.parent) + "/" + current.name;
        }

        public static T NewWithComponent<T>(string name) where T : Component
        {
            return (new GameObject(name, typeof(T))).GetComponent<T>();
        }

        public static T GetOrAddComponent<T>(this GameObject self) where T : Component
        {
            var component = self.GetComponent<T>();
            if (component == null)
                component = self.AddComponent<T>();
            return component;
        }

        public static T GetOrAddComponent<T>(this MonoBehaviour self) where T : Component
        {
            return self.gameObject.GetOrAddComponent<T>();
        }

        public static Vector2 xy(this Vector3 aVector) { return new Vector2(aVector.x, aVector.y); }
        public static Vector2 xz(this Vector3 aVector) { return new Vector2(aVector.x, aVector.z); }
        public static Vector2 yz(this Vector3 aVector) { return new Vector2(aVector.y, aVector.z); }
        public static Vector2 yx(this Vector3 aVector) { return new Vector2(aVector.y, aVector.x); }
        public static Vector2 zx(this Vector3 aVector) { return new Vector2(aVector.z, aVector.x); }
        public static Vector2 zy(this Vector3 aVector) { return new Vector2(aVector.z, aVector.y); }

        public static bool IsValid(this Plane plane)
        {
            return plane.normal.sqrMagnitude > 0.5f;
        }

        public static void SetKeywordEnabled(this Material mat, string name, bool enabled)
        {
            if (enabled) mat.EnableKeyword(name);
            else mat.DisableKeyword(name);
        }

        public struct FloatPackingPrecision { public const int High = 64; public const int Low = 8; public const int Undef = 0; }
        static int ms_FloatPackingPrecision = FloatPackingPrecision.Undef;

        // OpenGL ES 2.0 GPU (graphicsShaderLevel = 30) usually have low float precision (16 bits on fragments)
        // So we lower the float packing precision on them (8 seems fine on Adreno (TM) 220, NVIDIA Tegra 3 and on Mali-450 MP)
        // https://docs.unity3d.com/Manual/SL-DataTypesAndPrecision.html
        const int kFloatPackingHighMinShaderLevel = 35;

        public static int GetFloatPackingPrecision()
        {
            if (ms_FloatPackingPrecision == FloatPackingPrecision.Undef)
            {
                ms_FloatPackingPrecision = SystemInfo.graphicsShaderLevel >= kFloatPackingHighMinShaderLevel ? FloatPackingPrecision.High : FloatPackingPrecision.Low;
            }
            return ms_FloatPackingPrecision;
        }

        public static void LoadTexture(string url, TextureSettings settings, TextureSetCallback callback)
        {
            QueueFastLoadTexture(url, settings, callback);
        }

        private static void QueueFastLoadTexture(string url, TextureSettings settings, TextureSetCallback callback)
        {
            if (ImageLoaderThreaded.singleton == null)
                return;
            if (string.IsNullOrEmpty(url))
                return;

            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = false;
            queuedImage.skipCache = false;
            queuedImage.compress = settings.compress;
            queuedImage.createMipMaps = settings.createMipMaps;
            queuedImage.isNormalMap = settings.isNormalMap;
            queuedImage.linear = settings.linearColor;
            queuedImage.createAlphaFromGrayscale = settings.createAlphaFromGrayscale;
            queuedImage.createNormalFromBump = settings.createNormalFromBump;
            queuedImage.bumpStrength = settings.bumpStrength;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
            {
                Texture2D tex = qi.tex;
                if (tex != null)
                {
                    tex.wrapMode = settings.wrapMode;
                    tex.filterMode = settings.filterMode;
                    tex.anisoLevel = settings.anisoLevel;
                }
                callback(tex);
            };
            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }

        public static void SetLayer(Transform target, int layer, List<string> ignoreList = null)
        {
            if (ignoreList == null || !ignoreList.Contains(target.name))
                target.gameObject.layer = layer;

            if (target.childCount > 0)
            {
                for (int i = 0; i < target.childCount; i++)
                {
                    Transform child = target.GetChild(i);
                    SetLayer(child, layer, ignoreList);
                }
            }
        }

        public static Transform CenterCameraTrans
        {
            get
            {
                if (Utils.IsInVR())
                {
                    if (SuperController.singleton.isOpenVR)
                    {
                        return SuperController.singleton.ViveCenterCamera.transform;
                    }
                    return SuperController.singleton.OVRCenterCamera.transform;
                }
                else
                {
                    return SuperController.singleton.MonitorCenterCamera.transform;
                }
            }
        }

        public static bool IsInVR()
        {
            return SuperController.singleton.isOpenVR || SuperController.singleton.isOVR;
        }

        public static float GetMin(List<float> sourceList)
        {
            float min = sourceList[0];
            for (int x = 1; x < sourceList.Count; x++)
            {
                if (min > sourceList[x])
                {
                    min = sourceList[x];
                }
            }

            return min;
        }

        public static Transform GetScaleTransform(FreeControllerV3 controllerV3)
        {
            // scale target
            Transform tmpTrans;
            SetTransformScale setTransformScale = controllerV3.GetComponentInParent<SetTransformScale>();
            if (setTransformScale == null)
            {
                tmpTrans = controllerV3.transform.parent.Find("object/rescaleObject").transform;
            }
            else
            {
                tmpTrans = setTransformScale.transform;
            }

            return tmpTrans;
        }

        public static Atom GetAtomFromController(FreeControllerV3 controllerV3)
        {
            return controllerV3.transform.GetComponentInParent<Atom>();
        }

        public static Atom GetAtomFromInstanceID(int id)
        {
            return SuperController.singleton.GetAtoms().FirstOrDefault(x => x.GetInstanceID() == id);
        }

        public static FreeControllerV3 GetFreeControllerV3ByInstanceID(int id)
        {
            return SuperController.singleton.GetAllFreeControllers().FirstOrDefault(x => x.transform.GetInstanceID() == id);
        }

        public static FreeControllerV3 GetControllerFromAtom(Atom atom)
        {
            if (atom.mainController != null)
            {
                return atom.mainController;
            }
            else if (atom.containingAtom != null)
            {
                if (atom.containingAtom.mainController != null)
                {
                    return atom.containingAtom.mainController;
                }
            }

            return null;
        }

        public static Canvas AppendCanvasObjectUnderHeadCenter(string canvasName, bool isUICanInteractive)
        {
            Transform headCenter = SuperController.singleton.centerCameraTarget.transform;

            GameObject canvasObj = new GameObject();
            canvasObj.name = canvasName;
            var popupCanvas = canvasObj.AddComponent<Canvas>();
            popupCanvas.renderMode = RenderMode.WorldSpace;
            CanvasScaler cs = canvasObj.AddComponent<CanvasScaler>();
            cs.scaleFactor = 100.0f;
            cs.dynamicPixelsPerUnit = 4f;
            RectTransform rt = canvasObj.GetComponent<RectTransform>();
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, Screen.width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, Screen.height);
            canvasObj.transform.localScale = new Vector3(0.0006f, 0.0006f, 0.0006f);
            canvasObj.transform.localPosition = new Vector3(0.0f, 0.0f, .5f);

            if (isUICanInteractive)
            {
                canvasObj.AddComponent<GraphicRaycaster>();
                SuperController.singleton.AddCanvas(popupCanvas);
            }

            rt.SetParent(headCenter, false);

            return popupCanvas;
        }

        public static Transform GetAtomRescaleObject(Atom atom)
        {
            return atom.containingAtom.reParentObject.Find("object/rescaleObject");
        }

        public static Vector2 MapScreenPosTo01(Vector3 screenPosition)
        {
            return new Vector2(screenPosition.x / (float)Screen.currentResolution.width, screenPosition.y / (float)Screen.currentResolution.height);
        }

        public static float GetVectorAngleRange360(Vector3 from, Vector3 to)
        {
            // Determine whether the vector of the cross product points off the screen to determine whether the rotation angle is greater than 180.
            Vector3 crossVec = Vector3.Cross(from, to);
            float dotVal = Vector3.Dot(crossVec, Vector3.back);
            float angle = Vector3.Angle(from, to);

            if (dotVal < 0)
            {
                angle = 360 - angle;
            }
            return angle;
        }

        // The projection of src in the direction of target
        public static Vector3 VecProject(Vector3 src, Vector3 target)
        {
            return Vector3.Dot(src, target) * target.normalized;
        }

        // Point of intersection of X-ray and plane
        public static Vector3 GetIntersectPointWithRay(Vector3 planePoint, Vector3 n, Ray ray)
        {
            float d = Vector3.Dot(planePoint, n);
            Vector3 rayStartPoint = ray.origin;
            Vector3 rayDirction = ray.direction.normalized;
            float t0 = (d - Vector3.Dot(rayStartPoint, n)) / Vector3.Dot(rayDirction, n);
            return rayStartPoint + rayDirction * t0;
        }

        // Get the vertical vector of the two-dimensional vector (point multiplied by 0, there are two)
        public static Vector2 GetVerticalVectorFrom(Vector2 from)
        {
            return new Vector2(-from.y, from.x).normalized;
        }

        public static T GetSafeComponent<T>(GameObject obj) where T : Component
        {
            T comp = obj.GetComponent<T>();
            if (comp == null)
            {
                comp = obj.AddComponent<T>();
            }
            return comp;
        }

        // Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
        public static string GetPluginPath(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }

        public static JSONStorableBool CreateToggle(MVRScript self, string text, bool startValue, bool rightSide, bool acs)
        {
            JSONStorableBool storable = new JSONStorableBool(text, startValue);
            storable.storeType = JSONStorableParam.StoreType.Full;
            self.CreateToggle(storable, rightSide).label = Location.GetText(text);
            self.RegisterBool(storable);


            Toggle toggle = storable.toggle;
            toggle.enabled = acs;

            if (!acs)
                toggle.transform.Find("Panel").GetComponent<Image>().color = Color.gray;

            return storable;
        }

        // Create VaM-UI Float slider
        public static JSONStorableFloat CreateSliderFloat(MVRScript self, string text, float startValue, float min, float max, bool rightSide, bool acs)
        {
            JSONStorableFloat storable = new JSONStorableFloat(text, startValue, min, max, true, true);
            storable.storeType = JSONStorableParam.StoreType.Full;
            self.CreateSlider(storable, rightSide).label = Location.GetText(text);
            self.RegisterFloat(storable);

            Slider slider = storable.slider;
            slider.enabled = acs;

            if (!acs)
                slider.transform.parent.Find("Panel").GetComponent<Image>().color = Color.gray;

            return storable;
        }

        public static UIDynamicTwinButton SetupTwinButton(MVRScript script, string leftLabel, UnityAction leftCallback, string rightLabel, UnityAction rightCallback, bool rightSide, bool acs)
        {
            if (internalTwinButtonPrefab == null)
            {
                internalTwinButtonPrefab = new GameObject("TwinButton");
                internalTwinButtonPrefab.SetActive(false);
                RectTransform rt = internalTwinButtonPrefab.AddComponent<RectTransform>();
                rt.anchorMax = new Vector2(0, 1);
                rt.anchorMin = new Vector2(0, 1);
                rt.offsetMax = new Vector2(535, -500);
                rt.offsetMin = new Vector2(10, -600);
                LayoutElement le = internalTwinButtonPrefab.AddComponent<LayoutElement>();
                le.flexibleWidth = 1;
                le.minHeight = 50;
                le.minWidth = 350;
                le.preferredHeight = 50;
                le.preferredWidth = 500;

                RectTransform buttonTransform = script.manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
                buttonTransform = UnityEngine.Object.Instantiate(buttonTransform, internalTwinButtonPrefab.transform);
                buttonTransform.name = "ButtonLeft";
                buttonTransform.anchorMax = new Vector2(0.5f, 1.0f);
                buttonTransform.anchorMin = new Vector2(0.0f, 0.0f);
                buttonTransform.offsetMax = new Vector2(-3, 0);
                buttonTransform.offsetMin = new Vector2(0, 0);
                Button buttonLeft = buttonTransform.GetComponent<Button>();
                Text labelLeft = buttonTransform.Find("Text").GetComponent<Text>();

                buttonTransform = UnityEngine.Object.Instantiate(buttonTransform, internalTwinButtonPrefab.transform);
                buttonTransform.name = "ButtonRight";
                buttonTransform.anchorMax = new Vector2(1.0f, 1.0f);
                buttonTransform.anchorMin = new Vector2(0.5f, 0.0f);
                buttonTransform.offsetMax = new Vector2(0, 0);
                buttonTransform.offsetMin = new Vector2(3, 0);
                Button buttonRight = buttonTransform.GetComponent<Button>();
                Text labelRight = buttonTransform.Find("Text").GetComponent<Text>();

                UIDynamicTwinButton uid = internalTwinButtonPrefab.AddComponent<UIDynamicTwinButton>();
                uid.labelLeft = labelLeft;
                uid.labelRight = labelRight;
                uid.buttonLeft = buttonLeft;
                uid.buttonRight = buttonRight;

                uid.buttonLeft.enabled = acs;
                uid.buttonRight.enabled = acs;

                if (!acs)
                {
                    Image image = uid.buttonLeft.GetComponent<Image>();
                    image.color = Color.gray;

                    Image image2 = uid.buttonRight.GetComponent<Image>();
                    image2.color = Color.gray;
                }
            }

            {
                Transform t = internalCreateUIElement(internalTwinButtonPrefab.transform, rightSide);
                UIDynamicTwinButton uid = t.GetComponent<UIDynamicTwinButton>();
                uid.labelLeft.text = leftLabel;
                uid.labelRight.text = rightLabel;
                uid.buttonLeft.onClick.AddListener(leftCallback);
                uid.buttonRight.onClick.AddListener(rightCallback);
                t.gameObject.SetActive(true);

                uid.labelLeft.text = Location.GetText(leftLabel);
                uid.labelRight.text = Location.GetText(rightLabel);

                return uid;
            }
        }

        // Create VaM-UI Float slider
        public static JSONStorableFloat CreateSliderInt(MVRScript self, string text, int startValue, int min, int max, bool rightSide, bool acs)
        {
            JSONStorableFloat storable = new JSONStorableFloat(text, startValue, min, max, true, true);
            storable.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider slider = self.CreateSlider(storable, rightSide);
            slider.label = Location.GetText(text);
            slider.slider.wholeNumbers = true;
            slider.valueFormat = "F0";
            self.RegisterFloat(storable);

            Slider slider1 = storable.slider;
            slider1.enabled = acs;

            if (!acs)
                slider1.transform.parent.Find("Panel").GetComponent<Image>().color = Color.gray;

            return storable;
        }

        // Create VaM-UI ColorPicker
        public static JSONStorableColor CreateColorPicker(MVRScript self, string label, Color color, bool rightSide)
        {
            HSVColor hsvColor = HSVColorPicker.RGBToHSV(color.r, color.g, color.b);
            JSONStorableColor storable = new JSONStorableColor(label, hsvColor);
            storable.storeType = JSONStorableParam.StoreType.Full;
            self.CreateColorPicker(storable, rightSide).label = Location.GetText(label);
            self.RegisterColor(storable);
            return storable;
        }

        // Create VaM-UI StringChooser
        public static JSONStorableStringChooser CreateStringChooser(MVRScript self, string text, List<string> entries, bool rightSide, bool acs)
        {
            string defaultEntry = entries.Count > 0 ? entries[0] : "";
            JSONStorableStringChooser storable = new JSONStorableStringChooser(text, entries, defaultEntry, text);
            var popUI = self.CreateScrollablePopup(storable, rightSide);
            popUI.label = Location.GetText(text);
            self.RegisterStringChooser(storable);

            Image image = popUI.transform.Find("Button").GetComponentInChildren<Image>();
            image.raycastTarget = acs;

            if (!acs)
                image.color = Color.gray;

            return storable;
        }

        public static JSONStorableColor CreateColor(MVRScript self, string text, Color color, bool rightSide)
        {
            HSVColor hsvColor = HSVColorPicker.RGBToHSV(color.r, color.g, color.b);
            JSONStorableColor storable = new JSONStorableColor(text, hsvColor);
            storable.storeType = JSONStorableParam.StoreType.Full;
            var colorPicker = self.CreateColorPicker(storable, rightSide);
            colorPicker.label = Location.GetText(text);
            self.RegisterColor(storable);
            return storable;
        }

        public static JSONStorableUrl CreateFolderChooser(MVRScript self, string text, string defaultValue, bool rightSide, Action<string> callback, bool acs)
        {
            JSONStorableUrl storable = new JSONStorableUrl(text, string.Empty, (string url) => {
                if (callback != null) callback(GetFolderFromUrl(url));
            }, "png|jpg|jpeg|gif");
            self.RegisterUrl(storable);
            UIDynamicButton button = self.CreateButton("Browse " + text, rightSide);
            button.label = Location.GetText(text);

            SetupTextInput(self, "url", storable, false);
            if (!string.IsNullOrEmpty(defaultValue))
                storable.SetFilePath(defaultValue);
            storable.RegisterFileBrowseButton(button.button);

            button.button.enabled = acs;

            Image image = button.button.GetComponent<Image>();
            if (!acs)
                image.color = Color.gray;

            return storable;
        }

        public static JSONStorableUrl CreateFolderChooser(MVRScript self, string text, string defaultValue, bool rightSide, Action<string> callback, out UIDynamicButton browseBtn, out UIDynamicLabelInput urlInputText, bool acs)
        {
            JSONStorableUrl storable = new JSONStorableUrl(text, string.Empty, (string url) => {
                if (callback != null) callback(GetFolderFromUrl(url));
            }, "png|jpg|jpeg");
            self.RegisterUrl(storable);
            browseBtn = self.CreateButton("Browse " + text, rightSide);
            browseBtn.label = Location.GetText(text);

            urlInputText = SetupTextInput(self, "url", storable, true);
            if (!string.IsNullOrEmpty(defaultValue))
                storable.SetFilePath(defaultValue);
            storable.RegisterFileBrowseButton(browseBtn.button);

            browseBtn.button.enabled = acs;

            Image image = browseBtn.button.GetComponent<Image>();
            if (!acs)
                image.color = Color.gray;

            return storable;
        }

        public static JSONStorableUrl CreateTexture2DChooser(MVRScript self, string label, string defaultValue, bool rightSide, TextureSettings settings, TextureSetCallback callback, bool acs)
        {
            JSONStorableUrl storable = new JSONStorableUrl(label, string.Empty, (string url) => {
                if (url.EndsWith(".gif"))
                {
                    AnimatedGifPlayer lastFp = self.gameObject.GetComponent<AnimatedGifPlayer>();
                    if (lastFp != null) GameObject.Destroy(lastFp);
                    AnimatedGifPlayer fp = self.gameObject.AddComponent<AnimatedGifPlayer>();

                    fp.FileName = url;
                    fp.Loop = true;
                    fp.AutoPlay = true;
                    fp.Init();

                    fp.OnReady += () => { callback(fp.GifTexture); };
                }
                else
                {
                    QueueLoadTexture(url, settings, callback);
                }
            }, "jpg|png|tif|tiff|gif");
            self.RegisterUrl(storable);

            UIDynamicButton button = self.CreateButton("Browse " + label, rightSide);
            button.label = Location.GetText(label);
            button.button.enabled = acs;

            if (!acs)
            {
                Image image = button.button.GetComponent<Image>();
                image.color = Color.gray;
            }

            SetupTextInput(self, "url", storable, rightSide);
            if (!string.IsNullOrEmpty(defaultValue))
                storable.SetFilePath(defaultValue);
            storable.RegisterFileBrowseButton(button.button);
            return storable;
        }

        public static void QueueLoadTexture(string url, TextureSettings settings, TextureSetCallback callback)
        {
            if (string.IsNullOrEmpty(url))
                return;

            if (ImageLoaderThreaded.singleton == null)
                return;

            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = true;
            queuedImage.skipCache = true;
            queuedImage.compress = settings.compress;
            queuedImage.createMipMaps = settings.createMipMaps;
            queuedImage.isNormalMap = settings.isNormalMap;
            queuedImage.linear = settings.linearColor;
            queuedImage.createAlphaFromGrayscale = settings.createAlphaFromGrayscale;
            queuedImage.createNormalFromBump = settings.createNormalFromBump;
            queuedImage.bumpStrength = settings.bumpStrength;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
            {
                Texture2D tex = qi.tex;
                if (tex != null)
                {
                    tex.wrapMode = settings.wrapMode;
                    tex.filterMode = settings.filterMode;
                    tex.anisoLevel = settings.anisoLevel;
                }
                callback(tex);
            };
            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }

        // ".asf", ".avi", ".dv", ".m4v", ".mov", ".mp4", ".mpg", ".mpeg", ".ogv", ".vp8", ".webm", ".wmv"
        public static JSONStorableUrl SetupVideoChooser(MVRScript self, string text, string defaultValue, bool rightSide, JSONStorableString.SetStringCallback callback, bool acs)
        {
            JSONStorableUrl storable = new JSONStorableUrl(text, string.Empty, (string url) => {
                if (callback != null) callback(url);
            }, "mp4|avi|asf|dv|m4v|mov|mpg|mpeg|ogv|vp8|webm|wmv");
            self.RegisterUrl(storable);
            UIDynamicButton button = self.CreateButton("Browse " + text, rightSide);
            //button.label = Location.GetText(text);
            SetupTextInput(self, "url", storable, rightSide);
            storable.RegisterFileBrowseButton(button.button);

            button.button.enabled = acs;
            button.label = Location.GetText(text);
            Image image = button.button.GetComponent<Image>();
            if (!acs)
                image.color = Color.gray;

            return storable;
        }

        // Create VaM-UI StringChooser
        public static JSONStorableStringChooser CreateStringChooser(MVRScript self, string text, List<string> entries, int defaultIndex, bool rightSide, bool acs)
        {
            string defaultEntry = (defaultIndex >= 0 && defaultIndex < entries.Count) ? entries[defaultIndex] : "";
            JSONStorableStringChooser storable = new JSONStorableStringChooser(text, entries, defaultEntry, text);
            var popUI = self.CreateScrollablePopup(storable, rightSide);
            popUI.label = Location.GetText(text);
            self.RegisterStringChooser(storable);

            Image image = popUI.transform.Find("Button").GetComponentInChildren<Image>();
            image.raycastTarget = acs;

            if (!acs)
                image.color = Color.gray;

            return storable;
        }

        // Create VaM-UI Float slider
        public static JSONStorableFloat CreateSliderFloatWithRange(MVRScript script, string label, float defaultValue, float minValue, float maxValue, bool rightSide, bool acs)
        {
            JSONStorableFloat storable = new JSONStorableFloat(label, defaultValue, minValue, maxValue, true, true);
            storable.storeType = JSONStorableParam.StoreType.Full;
            storable.constrained = false;

            UIDynamicSlider dslider = script.CreateSlider(storable, rightSide);
            dslider.label = Location.GetText(label);
            Slider slider = dslider.slider;
            slider.enabled = acs;

            if (!acs)
                slider.transform.parent.Find("Panel").GetComponent<Image>().color = Color.gray;

            dslider.rangeAdjustEnabled = true;
            script.RegisterFloat(storable);
            return storable;
        }

        // Create VaM-UI InfoText field
        public static JSONStorableString CreateInfoText(MVRScript self, string text, float height, bool rightSide)
        {
            JSONStorableString storable = new JSONStorableString("Info", text);
            UIDynamic textfield = self.CreateTextField(storable, rightSide);
            textfield.height = height;
            return storable;
        }

        public static UIDynamic CreateSpacer(MVRScript script, float height, bool rightSide)
        {
            UIDynamic spacer = script.CreateSpacer(rightSide);
            spacer.height = height;
            return spacer;
        }

        // Create VaM-UI button
        public static UIDynamicButton CreateButton(MVRScript self, string text, UnityAction callback, bool rightSide, bool acs)
        {
            UIDynamicButton btn = self.CreateButton(text, rightSide);
            btn.button.onClick.AddListener(callback);
            btn.buttonText.text = Location.GetText(text);

            btn.button.enabled = acs;

            if (!acs)
            {
                Image image = btn.button.GetComponent<Image>();
                image.color = Color.gray;
            }

            return btn;
        }

        // Create input action trigger
        public static JSONStorableAction SetupAction(MVRScript self, string name, JSONStorableAction.ActionCallback callback)
        {
            JSONStorableAction action = new JSONStorableAction(name, callback);
            self.RegisterAction(action);
            return action;
        }

        // ===========================================================================================
        // Custom UI system with new UI elements and the ability to easily add/remove UI at runtime
        //
        // Usage instructions:
        // - Before using the custom UI elements, call from your MVRScript:
        //       Utils.OnInitUI(CreateUIElement);
        // - When your MVRScript receives the OnDestroy message call:
        //       Utils.OnDestroyUI();

        // Create one-line text input with label
        public static UIDynamicLabelInput SetupTextInput(MVRScript self, string text, JSONStorableString storable, bool rightSide)
        {
            if (internalInputPrefab == null)
            {
                CreateInternalInputPrefab(self);
            }

            Transform inputTrans = internalCreateUIElement(internalInputPrefab.transform, rightSide);
            UIDynamicLabelInput dinput = inputTrans.gameObject.GetComponent<UIDynamicLabelInput>();
            storable.inputField = dinput.input;
            dinput.text.text = Location.GetText(text);
            inputTrans.gameObject.SetActive(true);
            return dinput;
        }

        private static void CreateInternalInputPrefab(MVRScript self)
        {
            internalInputPrefab = new GameObject("LabelInput");
            internalInputPrefab.SetActive(false);
            RectTransform rt = internalInputPrefab.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0, 1);
            rt.anchorMin = new Vector2(0, 1);
            rt.offsetMax = new Vector2(535, -500);
            rt.offsetMin = new Vector2(10, -600);
            LayoutElement le = internalInputPrefab.AddComponent<LayoutElement>();
            le.flexibleWidth = 1;
            le.minHeight = 45;
            le.minWidth = 350;
            le.preferredHeight = 45;
            le.preferredWidth = 500;

            RectTransform backgroundTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
            backgroundTransform = UnityEngine.Object.Instantiate(backgroundTransform, internalInputPrefab.transform);
            backgroundTransform.name = "Background";
            backgroundTransform.anchorMax = new Vector2(1, 1);
            backgroundTransform.anchorMin = new Vector2(0, 0);
            backgroundTransform.offsetMax = new Vector2(0, 0);
            backgroundTransform.offsetMin = new Vector2(0, -10);

            RectTransform labelTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Button/Text") as RectTransform; ;
            labelTransform = UnityEngine.Object.Instantiate(labelTransform, internalInputPrefab.transform);
            labelTransform.name = "Text";
            labelTransform.anchorMax = new Vector2(0, 1);
            labelTransform.anchorMin = new Vector2(0, 0);
            labelTransform.offsetMax = new Vector2(155, -5);
            labelTransform.offsetMin = new Vector2(5, 0);
            Text labelText = labelTransform.GetComponent<Text>();
            labelText.text = "Name";
            labelText.color = Color.white;

            RectTransform inputTransform = self.manager.configurableTextFieldPrefab.transform as RectTransform;
            inputTransform = UnityEngine.Object.Instantiate(inputTransform, internalInputPrefab.transform);
            inputTransform.anchorMax = new Vector2(1, 1);
            inputTransform.anchorMin = new Vector2(0, 0);
            inputTransform.offsetMax = new Vector2(-5, -5);
            inputTransform.offsetMin = new Vector2(160, -5);
            UIDynamicTextField textfield = inputTransform.GetComponent<UIDynamicTextField>();
            textfield.backgroundColor = Color.white;
            LayoutElement layout = textfield.GetComponent<LayoutElement>();
            layout.preferredHeight = layout.minHeight = 35;
            InputField inputfield = textfield.gameObject.AddComponent<InputField>();
            inputfield.textComponent = textfield.UItext;
            inputfield.textComponent.resizeTextForBestFit = true;

            RectTransform textTransform = textfield.UItext.rectTransform;
            textTransform.anchorMax = new Vector2(1, 1);
            textTransform.anchorMin = new Vector2(0, 0);
            textTransform.offsetMax = new Vector2(-5, -5);
            textTransform.offsetMin = new Vector2(10, -5);

            UnityEngine.Object.Destroy(textfield);

            UIDynamicLabelInput uid = internalInputPrefab.AddComponent<UIDynamicLabelInput>();
            uid.text = labelText;
            uid.input = inputfield;
        }

        public static UIDynamicTextInfo CreateInfoTextNoScroll(MVRScript self, string text, float height, bool rightSide)
        {
            if (internalTextInfoPrefab == null)
            {
                CreateInternalTextInfoPrefab(self);
            }

            Transform textTrans = internalCreateUIElement(internalTextInfoPrefab.transform, rightSide);
            UIDynamicTextInfo dtext = textTrans.gameObject.GetComponent<UIDynamicTextInfo>();
            dtext.text.text = Location.GetText(text);
            dtext.layout.minHeight = height;
            dtext.layout.preferredHeight = height;
            textTrans.gameObject.SetActive(true);
            return dtext;
        }

        private static void CreateInternalTextInfoPrefab(MVRScript self)
        {
            internalTextInfoPrefab = new GameObject("TextInfo");
            internalTextInfoPrefab.SetActive(false);
            RectTransform rt = internalTextInfoPrefab.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0, 1);
            rt.anchorMin = new Vector2(0, 1);
            rt.offsetMax = new Vector2(535, -500);
            rt.offsetMin = new Vector2(10, -600);
            LayoutElement le = internalTextInfoPrefab.AddComponent<LayoutElement>();
            le.flexibleWidth = 1;
            le.minHeight = 35;
            le.minWidth = 350;
            le.preferredHeight = 35;
            le.preferredWidth = 500;

            RectTransform backgroundTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
            backgroundTransform = UnityEngine.Object.Instantiate(backgroundTransform, internalTextInfoPrefab.transform);
            backgroundTransform.name = "Background";
            backgroundTransform.anchorMax = new Vector2(1, 1);
            backgroundTransform.anchorMin = new Vector2(0, 0);
            backgroundTransform.offsetMax = new Vector2(0, 0);
            backgroundTransform.offsetMin = new Vector2(0, -10);

            RectTransform labelTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Button/Text") as RectTransform; ;
            labelTransform = UnityEngine.Object.Instantiate(labelTransform, internalTextInfoPrefab.transform);
            labelTransform.name = "Text";
            labelTransform.anchorMax = new Vector2(1, 1);
            labelTransform.anchorMin = new Vector2(0, 0);
            labelTransform.offsetMax = new Vector2(-5, 0);
            labelTransform.offsetMin = new Vector2(5, 0);
            Text labelText = labelTransform.GetComponent<Text>();
            labelText.alignment = TextAnchor.UpperLeft;

            UIDynamicTextInfo uid = internalTextInfoPrefab.AddComponent<UIDynamicTextInfo>();
            uid.text = labelText;
            uid.layout = le;
            uid.background = backgroundTransform;
        }

        public static UIDynamicTextInfo CreateInfoTextNoScroll(MVRScript self, JSONStorableString storable, float height, bool rightSide)
        {
            UIDynamicTextInfo uid = CreateInfoTextNoScroll(self, storable.val, height, rightSide);
            storable.setCallbackFunction = (string text) => {
                if (uid != null && uid.text != null)
                {
                    uid.text.text = Location.GetText(text);
                }
            };
            return uid;
        }

        public delegate Transform CreateUIElement(Transform prefab, bool rightSide);
        public static void InitUI(CreateUIElement createUIElementCallback)
        {
            internalCreateUIElement = createUIElementCallback;
        }

        public static void DestroyUI()
        {
            SafeDestroy(ref internalInputPrefab);
            SafeDestroy(ref internalTextInfoPrefab);
            SafeDestroy(ref internalTwinButtonPrefab);
        }

        private static void SafeDestroy(ref GameObject go)
        {
            if (go != null) GameObject.Destroy(go);
        }

        private static CreateUIElement internalCreateUIElement;
        private static GameObject internalInputPrefab;
        private static GameObject internalTextInfoPrefab;
        private static GameObject internalTwinButtonPrefab;
    }

    // ===========================================================================================

    public delegate void TextureSetCallback(Texture2D tex);

    public class TextureSettings
    {
        public bool compress = false;
        public bool createMipMaps = true;
        public bool isNormalMap = false;
        public bool linearColor = true; // Using linear or sRGB color space.
        public bool createAlphaFromGrayscale = false;
        public bool createNormalFromBump = false;
        public float bumpStrength = 1.0f;
        public TextureWrapMode wrapMode = TextureWrapMode.Repeat;
        public FilterMode filterMode = FilterMode.Trilinear;
        public int anisoLevel = 5; // 0: Forced off, 1: Off, quality setting can override, 2-9: Anisotropic filtering levels.
    }

    public class UIDynamicBase : UIDynamic
    {
    }

    public class UIDynamicLabelInput : UIDynamicBase
    {
        public Text text;
        public InputField input;
    }

    public class UIDynamicTextInfo : UIDynamicBase
    {
        public Text text;
        public LayoutElement layout;
        public RectTransform background;
    }

    public class UIDynamicTwinButton : UIDynamicBase
    {
        public Text labelLeft;
        public Text labelRight;
        public Button buttonLeft;
        public Button buttonRight;
    }
}
